//
//  GPU Normal DXT texture compression
//
//  This sample compresses DXT textures on the GPU using
//  integer textures and integer operations in the shader,
//  and the PBO extension for performing fast copies to a
//  DXT format texture.
//  
//  28 November, 2007
//  Thanks to Gary King, Eric Werness, Jan Paul van Waveren
//
// Author: Simon Green, Ignacio Castano
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.

#include <stdio.h>
#include <stdlib.h>
#include <map>
#include <GL/glew.h>
#include <GL/glut.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <time.h> // time
#include <math.h> // sin, cos

#include "nvImage.h"
#include "nvGlutManipulators.h"
#include "nvMath.h"
#include <nvSDKPath.h>
#include "framebufferObject.h"
#include "renderbuffer.h"

// Release 95 drivers have a bug with pixel buffer objects impacting this sample.
// The issue is fixed in release 100 and later drivers. Uncommenting the following
// line enables the use of PBOs.

#define USE_PBO
#define USE_GL_TIMER

using std::map;

enum UIOption {
    OPTION_DIFF,
    OPTION_SHADE,
    OPTION_COMPRESS,
    OPTION_LATC,
    OPTION_ANIMATE,
    OPTION_COUNT
};

bool options[OPTION_COUNT];
map<char, UIOption> optionKeyMap;

nv::GlutExamine manipulator;
int win_w = 512, win_h = 512;
int tex_w = win_w, tex_h = win_h;

nv::Image image;
int img_w, img_h;
GLuint tex;
GLuint compressed_dxt5_tex, compressed_latc_tex;
const char * img_filename = "media/textures/normalmap.png";

CGcontext cg_context; 
CGprofile cg_vprofile, cg_fprofile;
CGprogram compress_vprog, compress_dxt5_fprog, compress_latc_fprog;
CGprogram display_normal_fprog;
CGprogram display_object_vprog, display_object_fprog;

nv::SDKPath sdkPath;

FramebufferObject *scene_fbo, *compress_fbo;
GLuint scene_fbo_tex, compress_fbo_tex;
Renderbuffer * rb;

GLuint pbo; // pixel buffer object

float scale = 1.0;
float tx = 0.0, ty = 0.0;

void cgErrorCallback()
{
    CGerror lastError = cgGetError();
    if(lastError)
    {
        printf("%s\n", cgGetErrorString(lastError));
        printf("%s\n", cgGetLastListing(cg_context));
        exit(1);
    }
}

void setParameter(CGprogram prog, char *name, float x, float y=0.0, float z=0.0, float w=0.0)
{
    CGparameter param = cgGetNamedParameter(prog, name);
    if (param) {
        cgGLSetParameter4f(param, x, y, z, w);
    }
}

void setParameter(CGprogram prog, char *name, bool b)
{
    CGparameter param = cgGetNamedParameter(prog, name);
    if (param) {
        cgGLSetParameter1f(param, (float)b);
    }
}


void initCg()
{
    // create Cg Context
    cg_context = cgCreateContext();
    cgSetErrorCallback(cgErrorCallback);

    // load Cg programs
    cg_vprofile = cgGLGetLatestProfile(CG_GL_VERTEX);
    cg_fprofile = cgGLGetLatestProfile(CG_GL_FRAGMENT);

    const char *args[] = {
        "-unroll", "all",
        0,
    };

    std::string resolved_path;

    if (sdkPath.getFilePath( "src/compress_NormalDXT/compress_NormalDXT.cg", resolved_path)) {
        compress_vprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_vprofile, "compress_vp", args);
        cgGLLoadProgram(compress_vprog);

        compress_dxt5_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_fprofile, "compress_NormalDXT5_fp", args);
        cgGLLoadProgram(compress_dxt5_fprog);

        setParameter(compress_dxt5_fprog, "imageSize", tex_w, tex_h);

        compress_latc_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_fprofile, "compress_NormalLATC_fp", args);
        cgGLLoadProgram(compress_latc_fprog);

        setParameter(compress_latc_fprog, "imageSize", tex_w, tex_h);

        display_normal_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_fprofile, "display_fp", args);
        cgGLLoadProgram(display_normal_fprog);

        display_object_vprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_vprofile, "display_object_vp", args);
        cgGLLoadProgram(display_object_vprog);

        display_object_fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_fprofile, "display_object_fp", args);
        cgGLLoadProgram(display_object_fprog);
    }
    else {
        fprintf( stderr, "Failed to find shader file '%s'\n", "src/compress_NormalDXT/compress_NormalDXT.cg");
    }
}

GLuint createTexture(GLenum target, GLint internalformat, GLenum format, GLenum type, int w, int h)
{
    GLuint tex;
    glGenTextures(1, &tex);
    glBindTexture(target, tex);
    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(target, 0, internalformat, w, h, 0, format, type, 0);
    return tex;
}

void allocateBuffers(int w, int h)
{
    // size of compressed texture in bytes
    // Note that the algorithm relies on the texture dimensions being divisble by 4
    int size = (w/4)*(h/4)*16;

    GLubyte *data = new GLubyte [size];
    memset(data, 0, size);

    // create empty compressed textures
    glGenTextures(1, &compressed_dxt5_tex);
    glBindTexture(GL_TEXTURE_2D, compressed_dxt5_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glCompressedTexImage2DARB(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, w, h, 0, size, data);

    glGenTextures(1, &compressed_latc_tex);
    glBindTexture(GL_TEXTURE_2D, compressed_latc_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glCompressedTexImage2DARB(GL_TEXTURE_2D, 0, GL_COMPRESSED_LUMINANCE_ALPHA_LATC2_EXT, w, h, 0, size, data);

    delete [] data;

    // create fbo to render results of compression to
    compress_fbo = new FramebufferObject();
    compress_fbo_tex = createTexture(GL_TEXTURE_2D, GL_RGBA32UI_EXT, GL_RGBA_INTEGER_EXT, GL_INT, w/4, h/4);
    compress_fbo->Bind();
    compress_fbo->AttachTexture(GL_TEXTURE_2D, compress_fbo_tex, GL_COLOR_ATTACHMENT0_EXT);
    compress_fbo->IsValid();
    FramebufferObject::Disable();

    // scene fbo
    scene_fbo = new FramebufferObject();
    scene_fbo_tex = createTexture(GL_TEXTURE_2D, GL_RGBA8, GL_RGBA, GL_UNSIGNED_BYTE, w, h);
    rb = new Renderbuffer(GL_DEPTH_COMPONENT24, w, h);
    scene_fbo->AttachTexture(GL_TEXTURE_2D, scene_fbo_tex, GL_COLOR_ATTACHMENT0_EXT);
    scene_fbo->AttachRenderBuffer(rb->GetId(), GL_DEPTH_ATTACHMENT_EXT);
    scene_fbo->IsValid();
    FramebufferObject::Disable();

    // create PBO for copying integer data to compressed format
    glGenBuffersARB(1, &pbo);
    glBindBufferARB(GL_PIXEL_PACK_BUFFER_EXT, pbo);
    glBufferDataARB(GL_PIXEL_PACK_BUFFER_EXT, size, 0, GL_STREAM_COPY);
    glBindBufferARB(GL_PIXEL_PACK_BUFFER_EXT, 0);
}

void initOpengl()
{
    glewInit();

    if (!glewIsSupported(
        "GL_VERSION_2_0 "
        "GL_ARB_vertex_program "
        "GL_ARB_fragment_program "
        "GL_NV_gpu_program4 "
        "GL_ARB_pixel_buffer_object "
        "GL_EXT_framebuffer_object "
        "GL_ARB_texture_compression "
        "GL_EXT_texture_compression_s3tc "
        "GL_EXT_texture_compression_latc "
        "GL_EXT_texture_integer "
        ))
    {
        printf("Unable to load required extension\n");
        exit(-1);
    }

    glEnable(GL_DEPTH_TEST);
    glClearColor(0.2, 0.2, 0.2, 1.0);

    // load image
    std::string resolved_path;
    
    if (sdkPath.getFilePath( img_filename, resolved_path)) {
        image.loadImageFromFile(resolved_path.c_str());
        img_w = image.getWidth();
        img_h = image.getHeight();
        printf("Loaded '%s', %d x %d pixels\n", img_filename, img_w, img_h);
    }
    else {
        fprintf( stderr, "Failed to load image file '%s'\n", img_filename);
    }

    glGenTextures(1, &tex);
    glBindTexture(GL_TEXTURE_2D, tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, img_w, img_h, 0, image.getFormat(), image.getType(), image.getLevel(0));

    // object display list
    glNewList(1, GL_COMPILE);
    glutSolidTeapot(1.0);
    glEndList();

    allocateBuffers(tex_w, tex_h);

    glutReportErrors();
}

void drawQuad()
{
    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 0.0); glVertex2f(-1.0, -1.0);
    glTexCoord2f(1.0, 0.0); glVertex2f(1.0, -1.0);
    glTexCoord2f(1.0, 1.0); glVertex2f(1.0, 1.0);
    glTexCoord2f(0.0, 1.0); glVertex2f(-1.0, 1.0);
    glEnd();
}

// compress image by drawing a quad with compression shader
void compress()
{
    glViewport(0, 0, tex_w/4, tex_h/4);
    glDisable(GL_DEPTH_TEST);

    cgGLBindProgram(compress_vprog);
    cgGLEnableProfile(cg_vprofile);

    if (options[OPTION_LATC]) {
        cgGLBindProgram(compress_latc_fprog);
    }
    else {
        cgGLBindProgram(compress_dxt5_fprog);
    }
    
    cgGLEnableProfile(cg_fprofile);

    glBindTexture(GL_TEXTURE_2D, scene_fbo_tex);

    drawQuad();

    cgGLDisableProfile(cg_vprofile);
    cgGLDisableProfile(cg_fprofile);
}

// compress image and readback data to cpu
void compressReadback()
{
    // render to integer fbo
    compress_fbo->Bind();
    compress();

    // readback data to host
    int size = (tex_w/4)*(tex_h/4);
    GLushort *data = new GLushort [size*8];

    glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
    glReadPixels(0, 0, tex_w/4, tex_h/4, GL_RGBA_INTEGER_EXT, GL_UNSIGNED_INT, data);
    FramebufferObject::Disable();

    // download compressed texture
    if (options[OPTION_LATC])
    {
        glBindTexture(GL_TEXTURE_2D, compressed_latc_tex);
        glCompressedTexImage2DARB(GL_TEXTURE_2D, 0, GL_COMPRESSED_LUMINANCE_ALPHA_LATC2_EXT, tex_w, tex_h, 0, size*16, data);
    }
    else
    {
        glBindTexture(GL_TEXTURE_2D, compressed_dxt5_tex);
        glCompressedTexImage2DARB(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, tex_w, tex_h, 0, size*16, data);
    }

    delete [] data;
}

// compress image and use pbo to copy framebuffer to texture
void compressPBO()
{
    // render to integer fbo
    compress_fbo->Bind();
    compress();

    // readback int16 data to pbo
    glBindBufferARB(GL_PIXEL_PACK_BUFFER, pbo);
    glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
    glReadPixels(0, 0, tex_w/4, tex_h/4, GL_RGBA_INTEGER_EXT, GL_UNSIGNED_INT, 0);
    FramebufferObject::Disable();
    glBindBufferARB(GL_PIXEL_PACK_BUFFER, 0);

    // download to compressed texture from pbo
    // note - both these transfers will be implemented in the driver as fast copies in video memory
    int size = (tex_w/4)*(tex_h/4);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER, pbo);

    if (options[OPTION_LATC])
    {
        glBindTexture(GL_TEXTURE_2D, compressed_latc_tex);
        glCompressedTexImage2DARB(GL_TEXTURE_2D, 0, GL_COMPRESSED_LUMINANCE_ALPHA_LATC2_EXT, tex_w, tex_h, 0, size*16, 0);
    }
    else
    {
        glBindTexture(GL_TEXTURE_2D, compressed_dxt5_tex);
        glCompressedTexImage2DARB(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA_S3TC_DXT5_EXT, tex_w, tex_h, 0, size*16, 0);
    }

    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER, 0);
}


// benchmark compression
void benchmark(int iterations)
{
    glViewport(0, 0, tex_w/4, tex_h/4);
    glDisable(GL_DEPTH_TEST);

    cgGLBindProgram(compress_vprog);
    cgGLEnableProfile(cg_vprofile);

    if (options[OPTION_LATC])
    {
        cgGLBindProgram(compress_latc_fprog);
    }
    else
    {
        cgGLBindProgram(compress_dxt5_fprog);
    }

    cgGLEnableProfile(cg_fprofile);

    glBindTexture(GL_TEXTURE_2D, scene_fbo_tex);

#ifdef USE_GL_TIMER
    GLuint query;
    GLuint64EXT timeElapsed = 0;
    glGenQueries(1, &query);
    glBeginQuery(GL_TIME_ELAPSED_EXT, query);
#elif
    glFinish();
    int starttime = glutGet(GLUT_ELAPSED_TIME);
#endif

    for(int i=0; i<iterations; i++)
    {
        drawQuad();
        //compress();
    }

#ifdef USE_GL_TIMER
    glEndQuery(GL_TIME_ELAPSED_EXT);

    glGetQueryObjectui64vEXT(query, GL_QUERY_RESULT, &timeElapsed);
    glDeleteQueries(1, &query);
    float time = timeElapsed / (1e9);    // Nanosecons!
#else
    glFinish();
    int endtime = glutGet(GLUT_ELAPSED_TIME);
    float time = (endtime - starttime) / 1000.0;
#endif

    cgGLDisableProfile(cg_vprofile);
    cgGLDisableProfile(cg_fprofile);

    printf("%f secs, %f Mpixels/sec\n", time/iterations, (tex_w*tex_h*iterations) / (time*1e6));
}

// draw a simple scene
void drawScene()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // draw background
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

    glDisable(GL_DEPTH_TEST);
    glBindTexture(GL_TEXTURE_2D, tex);
    glEnable(GL_TEXTURE_2D);
    drawQuad();
    glDisable(GL_TEXTURE_2D);

    glPopMatrix();

    // draw teapot
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    manipulator.applyTransform();

    // Standard teapot transform.
    glRotated(270.0, 1.0, 0.0, 0.0);
    glScaled(0.5 * 1, -0.5 * 1, 0.5 * 1);
    glTranslated(0.0, 0.0, -1.5);

    cgGLBindProgram(display_object_vprog);
    cgGLEnableProfile(cg_vprofile);

    cgGLSetStateMatrixParameter(cgGetNamedParameter(display_object_vprog, "mvp"), 
        CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_TRANSPOSE);
    cgGLSetStateMatrixParameter(cgGetNamedParameter(display_object_vprog, "mvit"), 
        CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_INVERSE);

    cgGLBindProgram(display_object_fprog);
    cgGLEnableProfile(cg_fprofile);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glCallList(1);

    cgGLDisableProfile(cg_vprofile);
    cgGLDisableProfile(cg_fprofile);
}

// draw the scene to FBO
void drawSceneToTexture()
{
    scene_fbo->Bind();
    glViewport(0, 0, tex_w, tex_h);
    drawScene();
    FramebufferObject::Disable();
}

void glutPrint(float x, float y, const char *s, void *font)
{
    int i, len;

    glRasterPos2f(x, y);
    len = (int) strlen(s);
    for (i = 0; i < len; i++) {
        glutBitmapCharacter(font, s[i]);
    }
}

// display a given texture to the window
void displayTexture(GLuint tex)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glViewport(0, 0, win_w, win_h);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glScalef(scale, scale, 1.0);
    glTranslatef(tx, ty, 0.0);
    
    if (options[OPTION_COMPRESS] || options[OPTION_SHADE] || options[OPTION_DIFF])
    {
        cgGLBindProgram(display_normal_fprog);
        cgGLEnableProfile(cg_fprofile);

        if (options[OPTION_ANIMATE])
        {
            float t = clock() * 0.001f;
            setParameter(display_normal_fprog, "lightDirection", 0.7f * sinf(t), 0.7f * cosf(t), 1);
        }

        setParameter(display_normal_fprog, "displayNormal", !options[OPTION_SHADE]);
        setParameter(display_normal_fprog, "displayError", options[OPTION_DIFF]);
        setParameter(display_normal_fprog, "reconstructNormal", options[OPTION_COMPRESS]);
    }

    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_2D, scene_fbo_tex);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex);

    glEnable(GL_TEXTURE_2D);

    glDisable(GL_DEPTH_TEST);
    glColor3f(1.0, 1.0, 1.0);
    
    drawQuad();

    if (options[OPTION_COMPRESS] || options[OPTION_SHADE] || options[OPTION_DIFF])
    {
        cgGLDisableProfile(cg_fprofile);
    }

    glDisable(GL_TEXTURE_2D);

    glLoadIdentity();
    if (tex == compressed_dxt5_tex) {
        glutPrint(-0.95, -0.95, "DXT5 compressed", GLUT_BITMAP_9_BY_15);
    }
    else if (tex == compressed_latc_tex) {
        glutPrint(-0.95, -0.95, "LATC compressed", GLUT_BITMAP_9_BY_15);
    } 
    else {
        glutPrint(-0.95, -0.95, "original", GLUT_BITMAP_9_BY_15);
    }

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
}

void display()
{
    drawSceneToTexture();

    if ( options[OPTION_COMPRESS]) {
#ifdef USE_PBO
        compressPBO();
#else
        compressReadback();
#endif
        if ( options[OPTION_LATC]) {
            displayTexture(compressed_latc_tex);
        }
        else {
            displayTexture(compressed_dxt5_tex);
        }
    } else {
        displayTexture(scene_fbo_tex);
    }

    glutSwapBuffers();
}

void idle()
{
    if ( options[OPTION_ANIMATE]) {
        manipulator.idle();
        glutPostRedisplay();
    }
}

void key(unsigned char k, int x, int y)
{
    k = tolower(k);

    if (optionKeyMap.find(k) != optionKeyMap.end())
        options[optionKeyMap[k]] = ! options[optionKeyMap[k]];

    switch(k) {
    case 27:
    case 'q':
        exit(0);
        break;

    case '=':
    case '+':
        scale *= 1.5;
        break;

    case '-':
    case '_':
        scale /= 1.5;
        break;

    case 'r':
        scale = 1.0;
        tx = ty = 0.0;
        break;
    case 'b':
        benchmark(1000);
        break;
    }
    
    glutPostRedisplay();
}

void special(int key, int x, int y)
{
    switch(key) {
    case GLUT_KEY_RIGHT:
        tx -= 0.1;
        break;
    case GLUT_KEY_LEFT:
        tx += 0.1;
        break;
    case GLUT_KEY_DOWN:
        ty += 0.1;
        break;
    case GLUT_KEY_UP:
        ty -= 0.1;
        break;
    }
    glutPostRedisplay();
}


void resize(int w, int h)
{
    glViewport(0, 0, w, h);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();    
    gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 0.1, 10.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    manipulator.reshape(w, h);
    win_w = w; win_h = h;
}

void mouse(int button, int state, int x, int y)
{
    manipulator.mouse(button, state, x, y);
}

void motion(int x, int y)
{
    manipulator.motion(x, y);
}

void mainMenu(int i)
{
    key((unsigned char) i, 0, 0);
}

void initMenus()
{
    glutCreateMenu(mainMenu);
    glutAddMenuEntry("Toggle compression format [x]", 'x');
    glutAddMenuEntry("Toggle compressed/original image [c]", 'c');
    glutAddMenuEntry("Show difference [d]", 'd');
    glutAddMenuEntry("Toggle normal map shading [s]", 's');
    glutAddMenuEntry("Benchmark performance [b]", 'b');
    glutAddMenuEntry("Zoom in [+]", '+');
    glutAddMenuEntry("Zoom out [-]", '-');
    glutAddMenuEntry("Reset [r]", 'r');
    glutAddMenuEntry("Quit (esc)", '\033');
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitWindowSize(win_w, win_h);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
    glutCreateWindow("compress_NormalDXT");

    initOpengl();
    initCg();
    initMenus();


    optionKeyMap['d'] = OPTION_DIFF;
    options[OPTION_DIFF] = false;
    
    optionKeyMap['s'] = OPTION_SHADE;
    options[OPTION_SHADE] = false;

    optionKeyMap['c'] = OPTION_COMPRESS;
    options[OPTION_COMPRESS] = true;

    optionKeyMap['x'] = OPTION_LATC;
    options[OPTION_LATC] = true;

    optionKeyMap[' '] = OPTION_ANIMATE;
    options[OPTION_ANIMATE] = true;

    printf( "compress_NormalDXT - sample showing GPU texture compression\n");
    printf( "   q/[ESC]    - Quit the app\n");
    printf( "   [SPACE]    - Toggle animation on/off\n");
    printf( "      c       - Toggle compressing the texture on the GPU\n");
    printf( "      x       - Toggle compression format\n");
    printf( "      s       - Toggle normal map shading\n");
    printf( "      d       - Show difference\n");
    printf( "      r       - Reset the texture scale and translation\n");
    printf( "      b       - Benchmark the compression\n");
    printf( "      +       - Zoom into texture\n");
    printf( "      -       - Zoom out from texture\n");
    printf( "  [UP ARROW]  - Shift texture up\n");
    printf( " [DOWN ARROW] - Shift texture down\n");
    printf( " [LEFT ARROW] - Shift texture left\n");
    printf( " [RIGHT ARROW]- Shift texture right\n");
    printf( "   [MOUSE]    - Left - rotate, Left + Ctrl - dolly, Left + Shift - pan\n");
    printf( "   [MOUSE]    - Right - application menu\n\n");

    manipulator.setDollyActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_CTRL);
    manipulator.setPanActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_SHIFT);
    manipulator.setDollyPosition( -3.0f);

    glutDisplayFunc(display);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutSpecialFunc(special);
    glutReshapeFunc(resize);

    glutMainLoop();

    return 0;
}
